@@ -3,58 +3,64 @@ c     ~          ~~~   ~
 c     Clustering program based upon the k-medoid approach,
 c     and suitable for data sets of at least 100 objects.
 c     (for smaller data sets, please use program pam.)
-c     
+c
       subroutine clara(nn,jpp,kk,x,nran,nsam,dys,mdata,valmd,jtmd,ndyst,
      f     nrepr,nsel,nbest,nr,nrx,radus,ttd,ratt,ttbes,rdbes,rabes,
-     f     mtt,azba,avsyl,ttsyl,sylinf,jstop,
-     f     tmp1,tmp2,tmp3,ntmp1,ntmp2,ntmp3,ntmp4,ntmp5,ntmp6)
+     f     mtt,obj,avsyl,ttsyl,sylinf,jstop,
+     f     tmp1,tmp2,tmp3, ntmp1,ntmp2,ntmp3,ntmp4,ntmp5,ntmp6)
 
-      implicit double precision (a-h,o-z)
-      integer nn, jpp, nsam, nran, kk, mdata, jtmd, ndyst
+      integer nn, jpp, kk, nran, nsam, mdata, ndyst, jstop
 c     nn   = number of objects
 c     jpp  = number of variables
 c     kk   = number of clusters
-c     nran = ??
-c     nsam = number of objects drawn from data set
+c     nran = number of random samples drawn (= `samples' in S)
+c     nsam = number of objects drawn from data set (= `sampsize' in R)
+c
+c     mdata= {0,1};  1 : min(x) is missing value (NA);  0: no NA
+c     ndyst= {1,2};  1 : euclidean;  2 : manhattan
 
-c     The following vectors and matrices must be dimensioned in
-c     the main program:
-c     x(nn*jpp)
-c     jtmd(jpp),valmd(jpp)
-c     nrepr(nsam),nsel(nsam),nbest(nsam),dys(1 + nsam*(nsam-1)/2)
-c     nr(kk),nrx(kk),ttd(kk),radus(kk),ratt(kk)
-c     ttbes(kk),rdbes(kk),rabes(kk)
-c     
-      dimension x(nn*jpp)
-      dimension jtmd(jpp),valmd(jpp)
-      dimension nrepr(nsam),nsel(nsam),nbest(nsam)
-      dimension dys(1 + nsam*(nsam-1)/2)
-      dimension nr(kk),nrx(kk),ttd(kk),radus(kk),ratt(kk)
-      dimension ttbes(kk),rdbes(kk),rabes(kk),mtt(kk)
-      dimension sylinf(nsam,4),avsyl(kk)
-      dimension tmp1(nsam),tmp2(nsam),tmp3(nsam)
-      dimension ntmp1(nsam),ntmp2(nsam),ntmp3(nsam)
-      dimension ntmp4(nsam),ntmp5(nsam),ntmp6(nsam)
-      
-      integer*4 nrun
+c     valmd[j] = "missing value code" (instead of NA) for x[,j]
+c     jtmd [j] = {-1,1}:  -1 : x[,j] has NA ;  1 : no NAs in x[,j]
+
+      double precision x(nn*jpp), dys(1 + nsam*(nsam-1)/2),
+     1     valmd(jpp), radus(kk),ttd(kk),ratt(kk),
+     2     ttbes(kk),rdbes(kk),rabes(kk), obj, avsyl(kk), ttsyl,
+     3     sylinf(nsam,4), tmp1(nsam),tmp2(nsam),tmp3(nsam)
+      integer jtmd(jpp), nrepr(nsam),nsel(nsam),nbest(nsam),
+     1     nr(kk),nrx(kk), mtt(kk),
+     2     ntmp1(nsam),ntmp2(nsam),ntmp3(nsam),
+     3     ntmp4(nsam),ntmp5(nsam),ntmp6(nsam)
+c Var
+      logical nafs
+      integer j,jjb,jk,jkk,jn,js,jsm,jran,jhalt,
+     1     kall,kans,kkm,kkp,kran,
+     2     l,less, nad,nadv,nadvp,nexap,nexbp, nhalf,
+     3     nneq, nrun,nsamb,nsub, nunfs,nsm,ntt
+      double precision rnn, ran, zba, s,sx,z,zb
 
       jstop=0
       rnn=nn
-      nneq=0
-      if(nn.eq.nsam) nneq=1
+      if(nn.eq.nsam) then
+         nneq=1
+      else
+         nneq=0
+      endif
+c     nhalf := size of distance array dys()
       nhalf= nsam*(nsam-1)/2 + 1
       nsamb=2*nsam
-      nnpp=nn*jpp
+      if(nn.lt.nsamb) then
+         less=nn-nsam
+      else
+         less=nsam
+      endif
 
-c     
-c     in do 400, random subsamples are drawn and partitioned
-c     into kk clusters
-c     
- 120  nunfs=0
-      less=nsam
-      if(nn.lt.nsamb)less=nn-nsam
+      nunfs=0
       kall=0
+c     nrun : this is the ``random seed'' of the very simple randm() below
       nrun=0
+
+c
+c     Loop :  random subsamples are drawn and partitioned into kk clusters
       do 400 jran=1,nran
          jhalt=0
          if(nneq.eq.0)go to 140
@@ -105,8 +111,7 @@ C     Loop
  210     call randm(nrun,ran)
          kran=rnn*ran+1.
          if(kran.gt.nn)kran=nn
-         if(jran.eq.1)go to 230
-         if(nn.ge.nsamb)go to 230
+         if(jran.eq.1 .or. nn.ge.nsamb) go to 230
          do 220 jk=1,kk
             if(kran.eq.nrx(jk))go to 210
  220     continue
@@ -161,9 +166,11 @@ c     305  continue
          call bswap2(kk,nsam,nrepr,dys,z,s,tmp1,tmp2,tmp3)
          call selec(kk,nn,jpp,ndyst,zb,nsam,mdata,
      f        jtmd,valmd,nrepr,nsel,dys,x,nr,nafs,ttd,radus,ratt,
-     f        ntmp1,ntmp2,ntmp3,ntmp4,ntmp5,ntmp6,tmp1,tmp2)
-         nunfs=nunfs+nafs
-         if(nafs.eq.1)go to 400
+     f        ntmp1,ntmp2,ntmp3,ntmp4,ntmp5,ntmp6, tmp1,tmp2)
+         if (nafs) then
+            nunfs = nunfs+1
+            go to 400
+         endif
          if(jran.ne.1) then
             if(zb.ge.zba) go to 400
          endif
@@ -181,20 +188,21 @@ c     305  continue
  370     continue
          sx=s
  400  continue
+c--- end random sampling loop
 
       if(nunfs.ge.nran) then
          jstop=1
          return
       endif
-c     
+c
 c     for the best subsample, the objects of the entire data set
-c     are assigned to their clusters 
-c     
+c     are assigned to their clusters
+c
  450  if(kall.ne.1) then
          jstop=2
          return
       endif
- 460  azba=zba/rnn
+ 460  obj=zba/rnn
  470  call dysta2(nsam,jpp,nbest,x,nn,dys,ndyst,jtmd,valmd, jhalt)
       call resul(kk,nn,jpp,ndyst,mdata,jtmd,valmd, x,nrx,mtt)
       if(kk.gt.1) then
@@ -202,18 +210,16 @@ c
      +        ntmp1,ntmp2,ntmp3,ntmp4,tmp1,tmp2)
       endif
  500  end
-c     -----------------------------------------------------------
-c     
-c     
+c     --- of clara() ---------------------------------------------------
+c
+c
       subroutine dysta2(nsam,jpp,nsel,x,nn,dys,ndyst,jtmd, valmd,jhalt)
 
-      implicit none
-
       integer nsam,jpp, nn, nsel(nsam), ndyst, jtmd(jpp), jhalt
-      double precision x(nn*jpp), dys(1+nsam*(nsam-1)/2), valmd(jpp) 
+      double precision x(nn*jpp), dys(1+nsam*(nsam-1)/2), valmd(jpp)
 c
       double precision pp, rpres, clk
-      integer j,k,l, nlk, lsubt,lsel, ksel,numlj,numkj, npres
+      integer j,k,l, nlk, lsubt,lsel, ksel,lj,kj, npres
 c
       pp=jpp
       nlk=1
@@ -227,28 +233,30 @@ c
             nlk=nlk+1
             npres=0
             do 30 j=1,jpp
-               numlj=(lsel-1)*jpp+j
-               numkj=(ksel-1)*jpp+j
+               lj=(lsel-1)*jpp+j
+               kj=(ksel-1)*jpp+j
                if(jtmd(j).lt.0) then
-                  if(x(numlj).eq.valmd(j))go to 30
-                  if(x(numkj).eq.valmd(j))go to 30
+                  if(x(lj).eq.valmd(j))go to 30
+                  if(x(kj).eq.valmd(j))go to 30
                endif
  40            npres=npres+1
                if(ndyst.eq.1)then
-                  clk=clk+(x(numlj)-x(numkj))*(x(numlj)-x(numkj))
+                  clk=clk+(x(lj)-x(kj))*(x(lj)-x(kj))
                else
-                  clk=clk+dabs(x(numlj)-x(numkj))
+                  clk=clk+dabs(x(lj)-x(kj))
                endif
  30         continue
             rpres=npres
-            if(npres.ne.0)go to 60
-            jhalt=1
-            dys(nlk)=-1.0
-            go to 20
- 60         if(ndyst.ne.1)go to 70
-            dys(nlk)=dsqrt(clk*(pp/rpres))
-            go to 20
- 70         dys(nlk)=clk*(pp/rpres)
+            if(npres .eq. 0) then
+               jhalt=1
+               dys(nlk)=-1.0
+            else
+ 60            if(ndyst.eq.1) then
+                  dys(nlk)=dsqrt(clk*(pp/rpres))
+               else
+                  dys(nlk)=clk*(pp/rpres)
+               endif
+            endif
  20      continue
  100  continue
       end
@@ -259,8 +267,11 @@ c   we programmed this generator ourselves because we wanted it
 c   to be machine independent. it should run on most computers
 c   because the largest integer used is less than 2**30 . the period
 c   is 2**16=65536, which is good enough for our purposes.
-      implicit double precision (a-h,o-z)
-      integer*4 nrun,k
+      integer nrun
+      double precision ran
+
+      integer k
+      double precision ry
 
       nrun=nrun*5761+999
       k=nrun/65536
@@ -270,25 +281,32 @@ c   is 2**16=65536, which is good enough for our purposes.
       return
       end
 c     -----------------------------------------------------------
-c     
+c
       subroutine bswap2(kk,nsam,nrepr,dys,sky,s,dysma,dysmb,beter)
 
-      implicit double precision (a-h,o-z)
-      dimension dysma(nsam),dysmb(nsam),beter(nsam),nrepr(nsam)
-      dimension dys(1+nsam*(nsam-1)/2)
-c     
+      integer kk,nsam, nrepr(nsam)
+      double precision dys(1+nsam*(nsam-1)/2), sky,s,
+     1     dysma(nsam),dysmb(nsam),beter(nsam)
+c
+      integer meet
+      external meet
+c Var
+      integer j,k, ja, nny, njn,njaj,nkj,nmax,kbest,nbest
+      double precision ammax, rsam, asky,cmd,dz,dzsky,small
+
+c
 c     first algorithm: build.
-c     
+c
       nny=0
       do 17 j=1,nsam
          nrepr(j)=0
-         dysma(j)=1.1*s+1.0
+         dysma(j)= 1.1*s + 1.0
  17   continue
 
 c-- LOOP ---------------------
  20   do 22 ja=1,nsam
          if(nrepr(ja).ne.0)go to 22
-         
+
          beter(ja)=0.
          do 21 j=1,nsam
             njaj=meet(ja,j)
@@ -320,9 +338,9 @@ C--
       rsam=nsam
       asky=sky/rsam
 
-c     
+c
 c     second algorithm: swap.
-c     
+c
 C-- LOOP
  60   do 63 j=1,nsam
          dysma(j)=1.1*s+1.0
@@ -366,26 +384,38 @@ C-- LOOP
       sky=sky+dzsky
       go to 60
       end
+c     --- of bswap2() --------------------------------------------------
 
-c     -----------------------------------------------------------
-c     
+c     selec() : called once [per random sample] from clara()
+c
       subroutine selec(kk,nn,jpp,ndyst,zb,nsam,mdata,
      f     jtmd,valmd,nrepr,nsel,dys,x,nr,nafs,ttd,radus,ratt,
      f     nrnew,nsnew,npnew,ns,np,new,ttnew,rdnew)
 
-      implicit double precision (a-h,o-z)
-      dimension nrepr(nsam),nsel(nsam),dys(1+nsam*(nsam-1)/2)
-      dimension nrnew(nsam),nsnew(nsam),npnew(nsam),ttnew(nsam)
-      dimension rdnew(nsam)
-      dimension ns(nsam),nr(kk),np(nsam),ttd(kk),radus(kk),ratt(kk)
-      dimension jtmd(jpp),valmd(jpp),x(nn*jpp),new(nsam)
-c     
-c     nafs = 1 if a distance cannot be calculated
-c     
-      nafs=0
-c     
+      integer kk,nn,jpp,ndyst, nsam,mdata
+      integer jtmd(jpp), nrepr(nsam),nsel(nsam)
+      double precision zb, valmd(jpp), dys(1+nsam*(nsam-1)/2),
+     1     x(nn*jpp)
+      logical nafs
+      integer nr(kk)
+      double precision ttd(kk),radus(kk),ratt(kk),
+     1     ttnew(nsam), rdnew(nsam)
+      integer nrnew(nsam),nsnew(nsam),npnew(nsam),
+     1     ns(nsam),np(nsam), new(nsam)
+c
+      integer meet
+      external meet
+c Var
+      integer j,jk,jn,jp,jkabc,jnew, ka,kb, newf,
+     1     na,nb,nrjk, njk,npa,npb,npab,nstrt
+      double precision pp,dsum,dnull, tra, pres, abc, rns
+
+c
+c     nafs = .true. if a distance cannot be calculated
+      nafs=.false.
+c
 c     identification of representative objects, and initializations
-c     
+c
       jk=0
       do 10 j=1,nsam
          if(nrepr(j).eq.0)go to 10
@@ -396,11 +426,11 @@ c
          radus(jk)=-1.
          np(jk)=j
  10   continue
-c     
+c
 c     assignment of the objects of the entire data set to a cluster,
 c     computation of some statistics, determination of the
 c     new ordering of the clusters
-c     
+c
       zb=0.
       pp=jpp
       newf=0
@@ -459,7 +489,7 @@ c
             jkabc=jk
  70      continue
          if(pres.gt.0.5)go to 80
-         nafs=1
+         nafs=.true.
          return
       endif
 
@@ -479,10 +509,10 @@ c
       endif
  90   if(jn.lt.nn)go to 15
 
-c     
+c
 c     a permutation is carried out on vectors nr,ns,np,ttd,radus
 c     using the information in vector new.
-c     
+c
       do 92 jk=1,kk
          njk=new(jk)
          nrnew(jk)=nr(njk)
@@ -504,10 +534,10 @@ c
          ttd(j)=ttd(j)/rns
  101  continue
       if(kk.eq.1)go to 150
-c     
+c
 c     computation of minimal distance of medoid ka to any
 c     other medoid for comparison with the radius of cluster ka.
-c     
+c
       do 120 ka=1,kk
          nstrt=0
          npa=np(ka)
@@ -530,16 +560,20 @@ c
  150  return
       end
 c     -----------------------------------------------------------
-c     
-      subroutine resul(kk,nn,jpp,ndyst,mdata,jtmd,
-     f     valmd,x,nrx,mtt)
+c
+      subroutine resul(kk,nn,jpp,ndyst,mdata,jtmd, valmd,x,nrx,mtt)
+
+      integer kk,nn,jpp,ndyst,mdata
+      double precision x(nn*jpp), valmd(jpp)
+      integer jtmd(jpp), nrx(kk),mtt(kk)
+c Var
+      integer j,ja,jk,jn,jna,jksky, ka, na,nb,njnb,nrjk,nrjka,nxja
+      double precision pp,dsum,dnull,tra,abc
 
-      implicit double precision (a-h,o-z)
-      dimension x(nn*jpp),nrx(kk),jtmd(jpp),valmd(jpp),mtt(kk)
       pp=jpp
-c     
+c
 c     clustering vector is incorporated into x, and printed.
-c     
+c
       jn=0
  100  jn=jn+1
       njnb=(jn-1)*jpp
@@ -606,27 +640,34 @@ c
  320  continue
  330  end
 c     -----------------------------------------------------------
-c     
+c
       subroutine black(kk,jpp,nn,nsam,nbest,dys,sx,x,avsyl,ttsyl,sylinf,
      f     ncluv,nsend,nelem,negbr,syl,srank)
 
-      implicit double precision (a-h,o-z)
+      integer kk,jpp,nn, nsam, nbest(nsam)
+      double precision dys(1+nsam*(nsam-1)/2), sx, x(nn*jpp)
+      double precision avsyl(kk), ttsyl, sylinf(nsam,4)
+      integer ncluv(nsam),nsend(nsam),nelem(nsam),negbr(nsam)
+      double precision syl(nsam),srank(nsam)
+c
+      integer meet
+      external meet
+c Var
+      integer j,l,jna, lang,lplac, nbb,ncase,nsylr,numcl,nclu,ntt,nj,nl
+      double precision att,btt,rtt,db,dysa,dysb, rsam,symax
 
-      dimension ncluv(nsam),nsend(nsam),nelem(nsam),negbr(nsam)
-      dimension syl(nsam),srank(nsam),avsyl(kk),nbest(nsam)
-      dimension x(nn*jpp),dys(1+nsam*(nsam-1)/2),sylinf(nsam,4)
-c     
+c
 c     construction of clustering vector (ncluv)
 c     of selected sample (nbest).
-c     
+c
       do 12 l=1,nsam
          ncase=nbest(l)
          jna=(ncase-1)*jpp+1
          ncluv(l)=idint(x(jna)+0.1)
  12   continue
-c     
+c
 c     drawing of the silhouettes
-c     
+c
       nsylr=0
       ttsyl=0.0
       do 100 numcl=1,kk
@@ -648,8 +689,7 @@ c
                do 43 l=1,nsam
                   if(ncluv(l).ne.nclu)go to 43
                   nbb=nbb+1
-                  mjl=meet(nj,l)
-                  db=db+dys(mjl)
+                  db=db + dys(meet(nj,l))
  43            continue
                btt=nbb
                db=db/btt
@@ -661,8 +701,7 @@ c
             dysa=0.0
             do 45 l=1,ntt
                nl=nelem(l)
-               njl=meet(nj,nl)
-               dysa=dysa+dys(njl)
+               dysa=dysa + dys(meet(nj,nl))
  45         continue
             att=ntt-1
             dysa=dysa/att
